//! Shuttle service integration for the Poem web framework.
//! ## Example
//! ```rust,no_run
//! use poem::{get, handler, Route};
//! use shuttle_poem::ShuttlePoem;
//!
//! #[handler]
//! fn hello_world() -> &'static str {
//!     "Hello, world!"
//! }
//!
//! #[shuttle_runtime::main]
//! async fn poem() -> ShuttlePoem<impl poem::Endpoint> {
//!     let app = Route::new().at("/", get(hello_world));
//!
//!     Ok(app.into())
//! }
//!
//! ```

/// A wrapper type for [poem::Endpoint] so we can implement [shuttle_runtime::Service] for it.
pub struct PoemService<T>(pub T);

#[shuttle_runtime::async_trait]
impl<T> shuttle_runtime::Service for PoemService<T>
where
    T: poem::Endpoint + Send + 'static,
{
    async fn bind(mut self, addr: std::net::SocketAddr) -> Result<(), shuttle_runtime::Error> {
        poem::Server::new(poem::listener::TcpListener::bind(addr))
            .run(self.0)
            .await
            .map_err(shuttle_runtime::CustomError::new)?;

        Ok(())
    }
}

impl<T> From<T> for PoemService<T>
where
    T: poem::Endpoint + Send + 'static,
{
    fn from(router: T) -> Self {
        Self(router)
    }
}

/// Return type from the `[shuttle_runtime::main]` macro for a Poem-based service.
///
/// # Example
///
/// ```rust,no_run
/// use poem::{get, handler, Route};
/// use shuttle_poem::ShuttlePoem;
/// #[handler]
/// fn hello_world() -> &'static str {
///     "Hello, world!"
/// }
///
/// #[shuttle_runtime::main]
/// async fn poem() -> ShuttlePoem<impl poem::Endpoint> {
///     let app = Route::new().at("/", get(hello_world));
///
///     Ok(app.into())
/// }
/// ```
pub type ShuttlePoem<T> = Result<PoemService<T>, shuttle_runtime::Error>;
